/*
 * dummynet.c
 *
 *   Dummynet Streams Module Version 1.0
 *
 *   This is a streams module called "dummynet."  Rick Jones ported
 *   Luigi Rizzo's dummynet to a STREAMS module on HP-UX .  Yoshihiro
 *   Ishijima enhanced it so that the dummynet simulates various
 *   link characteristics such as frame error rate and frame error
 *   recovery at link level.
 *
 *             Copyright (C) 1999 Hewlett-Packard Company
 *                        ALL RIGHTS RESERVED.
 *
 * The enclosed software and documentation includes copyrighted works
 * of Hewlett-Packard Co. For as long as you comply with the following
 * limitations, you are hereby authorized to (i) use, reproduce, and
 * modify the software and documentation, and to (ii) distribute the
 * software and documentation, including modifications, for
 * non-commercial purposes only.
 *      
 * 1.  The enclosed software and documentation is made available at no
 *     charge in order to advance the general development of
 *     high-performance networking products.
 * 
 * 2.  You may not delete any copyright notices contained in the
 *     software or documentation. All hard copies, and copies in source
 *     code or object code form, of the software or documentation
 *     (including modifications) must contain at least one of the
 *     copyright notices.
 * 
 * 3.  The enclosed software and documentation has not been subjected
 *     to testing and quality control and is not a Hewlett-Packard Co.
 *     product. At a future time, Hewlett-Packard Co. may or may not
 *     offer a version of the software and documentation as a product.
 * 
 * 4.  THE SOFTWARE AND DOCUMENTATION IS PROVIDED "AS IS".
 *     HEWLETT-PACKARD COMPANY DOES NOT WARRANT THAT THE USE,
 *     REPRODUCTION, MODIFICATION OR DISTRIBUTION OF THE SOFTWARE OR
 *     DOCUMENTATION WILL NOT INFRINGE A THIRD PARTY'S INTELLECTUAL
 *     PROPERTY RIGHTS. HP DOES NOT WARRANT THAT THE SOFTWARE OR
 *     DOCUMENTATION IS ERROR FREE. HP DISCLAIMS ALL WARRANTIES,
 *     EXPRESS AND IMPLIED, WITH REGARD TO THE SOFTWARE AND THE
 *     DOCUMENTATION. HP SPECIFICALLY DISCLAIMS ALL WARRANTIES OF
 *     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *  
 * 5.  HEWLETT-PACKARD COMPANY WILL NOT IN ANY EVENT BE LIABLE FOR ANY
 *     DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
 *     (INCLUDING LOST PROFITS) RELATED TO ANY USE, REPRODUCTION,
 *     MODIFICATION, OR DISTRIBUTION OF THE SOFTWARE OR DOCUMENTATION.
 */

/*
 * The original dummynet code comes from Luigi Rizzo, and as per the
 * the terms of his copyright, his copyright statement is included
 * below. As for how to determine where Luigi ends and HP begins,
 * basically, if it is related to STREAMS, it is HP code. If it is as
 * the code appears in the original ip_dummynet.[ch], it is most
 * likely Luigi's...
 */

/*
 * Copyright (c) 1998 Luigi Rizzo
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * Redistribution in binary form may occur without any restrictions.
 * Obviously, it would be nice if you gave credit where credit is due
 * but requiring it would be too onerous.
 *
 * This software is provided ``AS IS'' without any warranties of any kind.
 *
 *      $Id: ip_dummynet.c,v 1.10 1999/03/26 14:15:59 luigi Exp $
 */

/*
 * Basic Theory of Operation for dummynet:
 *
 * This module implements the functionality of Luigi Rizzo's dummnet
 * tool. In a nutshell, we wrap a pushable STREAMS module around the
 * dummynet "pipes." Instead of getting to pipes by going through the
 * ipfw filtering of FreeBSD, we get here by having the module in the
 * stream. 
 *
 * The idea here is to wrap a pushable STREAMS module around the
 * "pipes" of the original dummynet implementation. Then, as a stream
 * is built, this module can be pushed into the apropriate place. A
 * common way to get this module in a stream under HP-UX 11 would be
 * to configure the autopush settings for /dev/tcp to include this
 * module before the tcpm module. Other STREAMS-based stacks would
 * likely be similar.  Another way to get this module in a stack 
 * would be to use ifconfig (what is the option name?  patch?)
 * to include the dummynet module between the link driver and ipm
 * module.
 *
 * Default settings for the pipes are stored in global variables
 * apropriately prefixed. These can be poked with adb or with
 * dummynet's ioctls commands.
 */


#if  defined(MODULEID) && !defined(lint)
static char rcsid[] = "@(#) dummynet.c experimental 1.00 1/25/2001";
#endif


/*
 * Turn this on if you want more debug info.
 */
#define DUMMYNET_DEBUG

#include <sys/types.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/stropts.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/malloc.h>
#include <sys/stream.h>
#include <sys/strenv.h>
#include <sys/tihdr.h>
#include <sys/dlpi.h>
#include <netinet/ip_info.h>
#include <sys/conf.h>

#include <mod_conf.h>
#include <sys/moddefs.h>

#include "dummynet.h"
#include "dummynetopt.h"

#ifndef TCPS_ESTABLISHED
/*
 * You need to install the latest Transport (TCP/IP) patch
 * that includes gettcpstate() function.  However, the patch
 * has not been released yet as of January 2001.
 *
 * I temporarilly compile dummynet with a hacked version of
 * gettcpstate() so that you can load the dummynet module without
 * the patch. Just use the mod.o that I provide.
 */
#include "gettcpstate.c"
#endif /* TCPS_ESTABLISHED */

/*
 * Functions
 */
int		dummynet_install();

/*
 * Static functions.
 */
static int	dummynet_load(void *);
static int	dummynet_unload(void *);
static int	dummynetopen(queue_t *, dev_t *, int, int);
static int	dummynetclose(queue_t *);
static int	dummynetrput(queue_t *, mblk_t *);
static int	dummynetwput(queue_t *, mblk_t *);
static void	dummynet_timeout(caddr_t);
static int	dummynet_ioctl(queue_t *, mblk_t *);
static void	dn_move(dn_pipe_t *, unsigned int);
static void	purge_pipe(dn_pipe_t *);
static void	dummynet_flush_putnext(dn_pipe_t *);
static void	dummynet_flush_freemsg(dn_pipe_t *);
static void	dummynet_piper(dn_pipe_t *, mblk_t *);
static void	dummynet_timeout_int(queue_t *, mblk_t *);
static int	dn_frame_rxmits(int, int, int, int, int *);
static void   	dn_compute_numbytes(dn_pipe_t *, int);
static int	dn_random();


/*
 * Streams data structs for "dummynet" module.
 */
static struct module_info dummynetm_info = {
    DUMMYNET_ID, /* module ID       */
    "dummynet",  /* name            */
    0,           /* min packet size */
    INFPSZ,      /* max packet size */
    0,           /* flow control's hi-water */
    0     	 /* flow control's lo-water */
};

static struct qinit dummynetrinit = {
    dummynetrput,    /* put routine     */
    NULL,	     /* service routine */
    dummynetopen,    /* open routine    */
    dummynetclose,   /* close routine   */
    NULL,            /* admin routine   */
    &dummynetm_info, /* module info     */
    NULL             /* module stat     */
};

static struct qinit dummynetwinit = {
    dummynetwput,    /* put routine     */
    NULL,	     /* service routine */
    NULL,            /* open routine    */
    NULL,            /* close routine   */
    NULL,            /* admin routine   */
    &dummynetm_info, /* module info     */
    NULL             /* module stat     */
};

struct streamtab dummynetinfo = {
    &dummynetrinit,
    &dummynetwinit };


/*
 * Dummynet module install routine and its data struct.
 *
 *    dummynet_install() is invoked by the I/O subsystem when it boots.
 *    (Someday, I'd like to make dummynet a dynamically loadable kernel
 *    module (DLKM).)
 */
static streams_info_t dummynet_str_info = {
    "dummynet",                       		/* name */
    -1,                         		/* major number */
    {&dummynetrinit,&dummynetwinit,NULL,NULL},	/* streamtab */
    STR_IS_MODULE | MGR_IS_MP,		  	/* streams flags */
    SQLVL_QUEUEPAIR,            		/* sync level */
    "",                         		/* elsewhere sync name */
};

/*
 * dummynet_instal()
 *
 *     This function is called if module is statically linked.
 */
int 
dummynet_install()
{
    int	retval;

    if ((retval = str_install (&dummynet_str_info)) != 0)
    {
	return (retval);
    }
    return 0;  /* return success */
}


/*
 * DLKM (Dynamically Loadable Kernel Module ) Wrapper Table
 */
extern struct mod_operations	str_mod_ops;
extern struct mod_conf_data	dummynet_conf_data; /* created by config() */

static struct mod_type_data	dummynet_drv_link = {
    "dummynet loadable streams module",
    (void *)&dummynet_str_info
};

static struct modlink		dummynet_mod_link[] = {
    {&str_mod_ops, (void *)&dummynet_drv_link},
    {NULL, (void *)NULL}
};

struct modwrapper		dummynet_wrapper = {
    MODREV,			/* module revision number */
    dummynet_load,		/* _load function	  */
    dummynet_unload,		/* _unload function	  */
    (void (*)())NULL,		/* currently not used	  */
    (void *)&dummynet_conf_data,/* configiguration data created by config() */
    dummynet_mod_link		/* modlink		  */
};


/*
 * dummynet_load
 */
static int
dummynet_load(void *arg)
{
    if (dummynet_install() != 0)
    {
	return ENXIO;
    }
    /*
     * Allocate module specific resources, if any.
     */
    return 0;
}

/*
 * dummynet_unload
 */
static int
dummynet_unload(void *arg)
{
    int		retval;
    
    retval = str_uninstall(&dummynet_str_info);
    if (retval != 0)
    {
	return retval;
    }
    /*
     * Dellocate module specific resources, if any.
     */
    return 0;
}


/*
 * External globals.
 */
extern int	ticks_since_boot;

/*
 * Dummynet's globals.
 */
int dn_placement                  = PUSH_DN_BETWEEN_TCP_UDP_AND_IP;

int dn_outbound_enable           = 0;
int dn_outbound_bandwidth_bps     =0; /* bits per second                    */
int dn_outbound_delay_ticks       =0; /* ticks                              */
int dn_outbound_queue_size        =0; /* # of packets: 1st choice           */
int dn_outbound_queue_size_bytes  =0; /* bytes:        2nd choice           */
int dn_outbound_frame_payload_size=0; /* frame's payload size (bytes)       */
int dn_outbound_ferror_per_million=0; /* # frame errors / one million frames*/
int dn_outbound_ferror_delay_ticks=0; /* cost of a local retransmit in ticks*/
int dn_outbound_ferror_rxmit_trials=0;/* # of attempts of local retransmits 
				       * excluding the original transmit    */
int dn_inbound_enable            = 0;
int dn_inbound_bandwidth_bps      =0; /* bits per second                    */
int dn_inbound_delay_ticks        =0; /* ticks                              */
int dn_inbound_queue_size         =0; /* # of packets: 1st choice           */
int dn_inbound_queue_size_bytes   =0; /* bytes:        2nd choice           */
int dn_inbound_frame_payload_size =0; /* frame's payload size (bytes)       */
int dn_inbound_ferror_per_million =0; /* # frame errors / one million frames*/
int dn_inbound_ferror_delay_ticks =0; /* cost of a local retransmit in ticks*/
int dn_inbound_ferror_rxmit_trials=0; /* # of attempts of local retransmits 
				       * excluding the original transmit    */


/*
 * Debug stuff
 */
#ifdef DUMMYNET_DEBUG
unsigned long	dn_stat0 = 0;
unsigned long	dn_stat1 = 0;
unsigned long	dn_stat2 = 0;
unsigned long	dn_stat3 = 0;
unsigned long	dn_stat4 = 0;
unsigned long	dn_stat5 = 0;
unsigned long	dn_stat6 = 0;
unsigned long	dn_stat7 = 0;
unsigned long	dn_stat8 = 0;
unsigned long	dn_stat9 = 0;
unsigned long	dn_stat10 = 0;
unsigned long	dn_stat11 = 0;
unsigned long	dn_stat12 = 0;
unsigned long	dn_stat13 = 0;
unsigned long	dn_stat14 = 0;
unsigned long	dn_stat15 = 0;
unsigned long	dn_stat16 = 0;
unsigned long	dn_stat17 = 0;
unsigned long	dn_stat18 = 0;
unsigned long	dn_stat19 = 0;
unsigned long	dn_stat20 = 0;
unsigned long	dn_stat21 = 0;
unsigned long	dn_stat22 = 0;
unsigned long	dn_stat23 = 0;
unsigned long	dn_stat24 = 0;
unsigned long	dn_stat25 = 0;
unsigned long	dn_stat26 = 0;
unsigned long	dn_stat27 = 0;
unsigned long	dn_stat28 = 0;
unsigned long	dn_stat29 = 0;
unsigned long	dn_stat30 = 0;
unsigned long	dn_stat31 = 0;
unsigned long	dn_stat32 = 0;
unsigned long	dn_stat33 = 0;
unsigned long	dn_stat34 = 0;
unsigned long	dn_stat35 = 0;
unsigned long	dn_stat36 = 0;
unsigned long	dn_stat37 = 0;
unsigned long	dn_stat38 = 0;
unsigned long	dn_stat39 = 0;
unsigned long	dn_stat40 = 0;
unsigned long	dn_stat41 = 0;
unsigned long	dn_stat42 = 0;
unsigned long	dn_stat43 = 0;
unsigned long	dn_stat44 = 0;
unsigned long	dn_stat45 = 0;
unsigned long	dn_stat46 = 0;
unsigned long	dn_stat47 = 0;
unsigned long	dn_stat48 = 0;
unsigned long	dn_stat49 = 0;
unsigned long	dn_stat50 = 0;
unsigned long	dn_stat51 = 0;
unsigned long	dn_stat52 = 0;
unsigned long	dn_stat53 = 0;
unsigned long	dn_stat54 = 0;
unsigned long	dn_stat55 = 0;
unsigned long	dn_stat56 = 0;
unsigned long	dn_stat57 = 0;
unsigned long	dn_stat58 = 0;
unsigned long	dn_stat59 = 0;
unsigned long	dn_stat60 = 0;
unsigned long	dn_stat61 = 0;
unsigned long	dn_stat62 = 0;
unsigned long	dn_stat63 = 0;
unsigned long	dn_stat64 = 0;
unsigned long	dn_stat65 = 0;
unsigned long	dn_stat66 = 0;
unsigned long	dn_stat67 = 0;
unsigned long	dn_stat68 = 0;
unsigned long	dn_stat69 = 0;
unsigned long	dn_stat70 = 0;
unsigned long	dn_stat71 = 0;
unsigned long	dn_stat72 = 0;
unsigned long	dn_stat73 = 0;
unsigned long	dn_stat74 = 0;
unsigned long	dn_stat75 = 0;
unsigned long	dn_stat76 = 0;
unsigned long	dn_stat77 = 0;
unsigned long	dn_stat78 = 0;
unsigned long	dn_stat79 = 0;
#endif /* DUMMYNET_DEBUG */

#ifdef DUMMYNET_DEBUG
#define DN_STAT_INCR(x)	(x)++
#else /* DUMMYNET_DEBUG */
#define DN_STAT_INCR(x)
#endif /* DUMMYNET_DEBUG */


/*
 * dummynetopen()
 *
 *    This is called when dummynet module instance is pushed onto a stream.
 *    We need to allocate pipes and the like and will put references
 *    to them in the structure pointed to by the queue private
 *    pointer - q->q_ptr;
 */
static int
dummynetopen(queue_t *q, dev_t *dev, int flag, int sflag)
{
    int		i;
    dn_d_t	*d;
    dn_pipe_t	*pipe;

    /*
     * Allocate a dummynet descriptor: dn_d.
     */
    d = (struct dn_d *) kmalloc (sizeof(struct dn_d), M_STREAMS, M_WAITOK);
    if (d == NULL)
    {
	return ENOMEM;
    }
    WR(q)->q_ptr = q->q_ptr = d;
    /*
     * Initialize dn_flags.
     */
    d->dn_flags = 0;
    if (dn_placement == PUSH_DN_BETWEEN_TCP_UDP_AND_IP)
    {
	d->dn_flags |= DN_BETWEEN_TCP_UDP_AND_IP;
    }
    else if (dn_placement == PUSH_DN_BETWEEN_IP_AND_DRIVER)
    {
	d->dn_flags |= DN_BETWEEN_IP_AND_DRIVER;
    }
    /*
     * Timer stuff.   Allocate a dummy mblk that is used for calling
     * streams_put from dummynet_timeout.
     */
    if((d->dn_timer_mp = allocb(1, BPRI_LO)) == NULL)
    {
	kfree ((caddr_t) d, M_STREAMS);
	return ENOMEM;
    }
    d->dn_timer_id            = 0;
    d->dn_last_ticks          = ticks_since_boot;
    /*
     * Initialize outbound pipe.
     */
    pipe = &d->dn_pipe_outbound;
    pipe->pipe_flags          = dn_outbound_enable ? DN_PIPE_ENABLED : 0;
    pipe->pipe_id             = 0;
    pipe->owner_q             = WR(q);
    pipe->routerq_head        = NULL;
    pipe->routerq_tail	      = NULL;
    pipe->routerq_len         = 0;
    pipe->routerq_len_bytes   = 0;
    pipe->routerq_size        = dn_outbound_queue_size;
    pipe->routerq_size_bytes  = dn_outbound_queue_size_bytes;
    pipe->routerq_drops       = 0;
    pipe->router_q_hiwat      = 0;
    pipe->bandwidth_bps       = dn_outbound_bandwidth_bps;
    pipe->numbytes            = 0;
    pipe->bytes_borrowed      = 0;
    pipe->pipeq_head	      = NULL;
    pipe->pipeq_tail	      = NULL;
    pipe->delay_ticks         = dn_outbound_delay_ticks;
    pipe->frame_payload_size  = dn_outbound_frame_payload_size;
    pipe->ferror_per_million  = dn_outbound_ferror_per_million;
    pipe->ferror_delay_ticks  = dn_outbound_ferror_delay_ticks;
    pipe->ferror_rxmit_trials = dn_outbound_ferror_rxmit_trials;
    /*
     * Initialize inbound pipe.
     */
    pipe = &d->dn_pipe_inbound;
    pipe->pipe_flags          = dn_inbound_enable ? DN_PIPE_ENABLED : 0;
    pipe->pipe_id             = 0;
    pipe->owner_q             = q;
    pipe->routerq_head        = NULL;
    pipe->routerq_tail	      = NULL;
    pipe->routerq_len         = 0;
    pipe->routerq_len_bytes   = 0;
    pipe->routerq_size        = dn_inbound_queue_size;
    pipe->routerq_size_bytes  = dn_inbound_queue_size_bytes;
    pipe->routerq_drops       = 0;
    pipe->router_q_hiwat      = 0;
    pipe->bandwidth_bps       = dn_inbound_bandwidth_bps;
    pipe->numbytes            = 0;
    pipe->bytes_borrowed      = 0;
    pipe->pipeq_head	      = NULL;
    pipe->pipeq_tail	      = NULL;
    pipe->delay_ticks         = dn_inbound_delay_ticks;
    pipe->frame_payload_size  = dn_inbound_frame_payload_size;
    pipe->ferror_per_million  = dn_inbound_ferror_per_million;
    pipe->ferror_delay_ticks  = dn_inbound_ferror_delay_ticks;
    pipe->ferror_rxmit_trials = dn_inbound_ferror_rxmit_trials;

    DN_STAT_INCR(dn_stat0);
    return 0;
}


/*
 * dummynetclose()
 */
static int
dummynetclose(queue_t *q)
{
    dn_d_t		*d = q->q_ptr;

    /*
     * 	Set a flag so that we will not post any more new timeouts.
     */
    d->dn_flags |= DN_CLOSING;
    /*
     * Get rid of our timer if it is running.
     */
    if (d->dn_timer_id)
    {
	untimeout(d->dn_timer_id);
    }
    d->dn_timer_id = 0;
    if (d->dn_timer_mp)
    {
	freemsg(d->dn_timer_mp);
	d->dn_timer_mp = 0;
    }
    /*
     * Flush the packets queued in the pipes, if any.
     */
    dummynet_flush_putnext(&d->dn_pipe_inbound);
    dummynet_flush_putnext(&d->dn_pipe_outbound);
    /*
     * Free the dummynet descriptor.
     */
    kfree ((caddr_t) d, M_STREAMS);

    DN_STAT_INCR(dn_stat1);
    return 0;
}


/*
 * dummynetrput()
 * 
 *    Read-side (inbound side) of the dummynet.
 */
static int
dummynetrput (queue_t *q, mblk_t *bp)
{
    dn_d_t		*d;
    dn_pipe_t		*pipe;
    int			len;
    int			tcp_state;
    int			type;
    unsigned int	dn_flags;
    unsigned int	pipe_flags;

    d          = (dn_d_t *) q->q_ptr;
    dn_flags   = d->dn_flags;
    pipe       = &d->dn_pipe_inbound;
    pipe_flags = pipe->pipe_flags;
    
    if (!(pipe_flags & DN_PIPE_ENABLED))
    {
	putnext(q, bp);
	return;
    }
    type = bp->b_datap->db_type;
    /*
     * Handle M_FLUSH.
     */
    if (type == M_FLUSH)
    {
	DN_STAT_INCR(dn_stat2);
	if (*bp->b_rptr & FLUSHR)
	{
	    dummynet_flush_freemsg(&d->dn_pipe_inbound);
	}
	if (*bp->b_rptr & FLUSHW)
	{
	    dummynet_flush_freemsg(&d->dn_pipe_outbound);
	}
	putnext(q, bp);
	return;
    }

    if (type >= QPCTL)
    {
	putnext(q, bp);		/* Just pass the priority messages. */
	return;
    }

    if (dn_flags & DN_BETWEEN_TCP_UDP_AND_IP)
    {
	/*
	 * Dummynet is pushed between tcp and ip.
	 *
	 * If the connection is not established yet,
	 * just pass the message.  This is because there
	 * may be a race contition for passive open case
	 * like the following:
	 *
	 * When a listener gets a SYN from a remote, TCP sends
	 * T_CONN_IND upstream.  Before the application accepts the
	 * connection, if the ack for SYN|ACK (and the first
	 * couple of segments) comes in, the ack is routed toward
	 * the dummynet under the *listener* TCP.  The dummynet
	 * would delay (bandwidth-control) sending the segment
	 * upstream, and meanwhile the application completes
	 * accepting the connection, and have its own
	 * tcp/dummynet/ip stack.  A little later, when the
	 * dummynet under the listen TCP eventually tries to
	 * send the pending segment(s) upstream, however, it sends
	 * them to the listening stack instead of the accepting
	 * stack, thus the segments may be dropped.
	 *
	 * In order to avoid the above scenario, when the upper
	 * TCP has not established a connection yet, we just pass
	 * the segment upstream.
	 *
	 * There is a drawback of doing this: until the connection
	 * established, dummynet will not control the bandwidth for
	 * inbound segments.  However, dummynet will still control
	 * the bandwidth and latency of the outbound segments.
	 */
	tcp_state = gettcpstate(q->q_next);
	if (tcp_state == TCPS_ESTABLISHED ||
	    tcp_state == TCPS_FIN_WAIT_1  ||
	    tcp_state == TCPS_FIN_WAIT_2)
	{
	    dummynet_piper(&d->dn_pipe_inbound, bp);
	    return;
	}
	else if (tcp_state == TCPS_INVALID)
	{
	    /*
	     * Dummynet is not pushed between TCP and IP.
	     * It is probably pushed between UDP and IP.
	     * Just call dummynet_piper.
	     */
	    dummynet_piper(&d->dn_pipe_inbound, bp);
	    return;
	}
	else
	{
	    DN_STAT_INCR(dn_stat3);

	    dummynet_flush_putnext(&d->dn_pipe_inbound);
	    putnext(q, bp);
	    return;
	}
    }
    else if (dn_flags & DN_BETWEEN_IP_AND_DRIVER)
    {
	/*
	 * Dummynet is pushed between ip and driver.
	 */
	dummynet_piper(&d->dn_pipe_inbound, bp);
	return;
    }
    else
    {
	/*
	 * We should not reach here.
	 */
	DN_STAT_INCR(dn_stat4);
	putnext(q, bp);
	return;
    }
}


/* 
 * dummynetwput()
 *
 *    Write-side (outbound side) of the dummynet.
 */
static int
dummynetwput(queue_t *q, mblk_t *bp)
{
    dn_d_t	 	*d;
    dn_pipe_t		*pipe;
    int			len;
    int			type;
    int			tcp_state;
    unsigned int	dn_flags;
    unsigned int	pipe_flags;

    d          = q->q_ptr;
    dn_flags   = d->dn_flags;
    pipe       = &d->dn_pipe_outbound;
    pipe_flags = pipe->pipe_flags;
    
    type = bp->b_datap->db_type;
    switch (type)
    {
	case M_FLUSH:
	{
	    DN_STAT_INCR(dn_stat5);
	    if (*bp->b_rptr & FLUSHR)
	    {
		dummynet_flush_freemsg(&d->dn_pipe_inbound);
	    }
	    if (*bp->b_rptr & FLUSHW)
	    {
		dummynet_flush_freemsg(&d->dn_pipe_outbound);
	    }
	    putnext(q, bp);
	    return;
	}
	case M_IOCTL:
	{
	    DN_STAT_INCR(dn_stat6);
	    if (dummynet_ioctl(q, bp))
	    {
		return;
	    }
	    break; /* M_IOCTL was not for the dummynet. Pass it downstream. */
	}
    }

    if (!(pipe_flags & DN_PIPE_ENABLED))
    {
	putnext(q, bp);
	return;
    }

    if (type >= QPCTL)
    {
	putnext(q, bp);		/* Just pass the priority messages. */
	return;
    }

    dummynet_piper(&d->dn_pipe_outbound, bp);
    return;
}


/*
 * dummynet_piper()
 *
 *    Dummynet has a pair of pipes: one for outbound (write-side,
 *    sending-side), and one for inbound (read-side, receiving-side).
 *    Each pipe has a router queue and a pipe queue.  The router
 *    queue controls the bandwidth of the link simulated by the
 *    dummynet.  The pipe queue controls the latency of th link
 *    simulated by the dummynet.  Each packet is first queued in
 *    the router queue, and waits for the sufficient bandwidth
 *    become available for sending the packet.  Then, the packet
 *    is moved to the pipe queue, and waits for the delay/latency
 *    time to be passed.  The delay time simulates the latency
 *    between the sender and receiver over the simulated link,
 *    and it includes the processing time of the transmiter and
 *    the receiver, the propagation delay, the processing time
 *    of link-levle frame retransmission(s), etc.
 */
static void
dummynet_piper(dn_pipe_t *pipe, mblk_t *mp)
{
    queue_t		*q;
    dn_d_t		*d;
    unsigned char	type;
    unsigned long	len;
    unsigned long	xmit_len;
    union T_primitives	*tpip;
    union DL_primitives	*dlpip;
    int			primitive;
    int			local_rxmits;
    unsigned int	current_ticks;
    unsigned int	dn_flags;
    
    DN_STAT_INCR(dn_stat7);

    q        = pipe->owner_q;
    d        = q->q_ptr;
    dn_flags = d->dn_flags;

#ifdef DUMMYNET_DEBUG
    if (mp->b_next)
    {
	panic("dummynet_piper: mp->b_next is not NULL.\n");
    }
#endif /* DUMMYNET_DEBUG */

    type = mp->b_datap->db_type;
    /*
     * Find the data length.  If it's not a data mblk, set the length to 0.
     */
    if (type == M_DATA)
    {
	DN_STAT_INCR(dn_stat8);
	if (!mp->b_cont)
	{
	    DN_STAT_INCR(dn_stat9);
	    len = mp->b_wptr - mp->b_rptr;
	}
	else
	{
	    DN_STAT_INCR(dn_stat10);
	    len = msgdsize(mp);
	}
    }
    else if (type == M_PROTO)
    {
	DN_STAT_INCR(dn_stat11);
	if (dn_flags & DN_BETWEEN_IP_AND_DRIVER)
	{
	    /*
	     * dummynet module is pushed between driver and ip.
	     */
	    dlpip = (union DL_primitives *) mp->b_rptr;
	    primitive = dlpip->dl_primitive;
	    if (primitive == DL_UNITDATA_REQ || primitive == DL_UNITDATA_IND)
	    {
		DN_STAT_INCR(dn_stat12);
		if (!mp->b_cont)
		{
		    /*
		     * No chained M_DATA.
		     */
		    DN_STAT_INCR(dn_stat13);
		    len = 0;
		}
		else
		{
		    DN_STAT_INCR(dn_stat14);
		    len = msgdsize(mp->b_cont);
		}
	    }
	    else
	    {
		DN_STAT_INCR(dn_stat15);
		len = 0;
	    }
	}
	else if (dn_flags & DN_BETWEEN_TCP_UDP_AND_IP)
	{
	    /*
	     * dummynet module is pushed between tcp/udp and ip.
	     */
	    tpip = (union T_primitives *) mp->b_rptr;
	    primitive = tpip->type;
	    if (primitive == T_DATA_REQ || primitive == T_DATA_IND ||
		primitive == T_UNITDATA_REQ || primitive == T_UNITDATA_IND)
	    {
		DN_STAT_INCR(dn_stat16);
		if (!mp->b_cont)
		{
		    /*
		     * No chained M_DATA.
		     */
		    DN_STAT_INCR(dn_stat17);
		    len = 0;
		}
		else
		{
		    DN_STAT_INCR(dn_stat18);
		    len = msgdsize(mp->b_cont);
		}
	    }
	    else
	    {
		DN_STAT_INCR(dn_stat19);
		len = 0;
	    }
	}
	else
	{
	    /*
	     * We should not reach here.
	     */
	    len = 0;
	}
    }
    else
    {
	DN_STAT_INCR(dn_stat20);
	len = 0;
    }
    if (len == 0)
    {
	DN_STAT_INCR(dn_stat21);
	xmit_len             = 0;
	*DN_PKT_LEN(mp)      = 0;
	*DN_PKT_XMIT_LEN(mp) = 0;
	*DN_PKT_DELAY(mp)    = 0;
	*DN_PKT_ERROR(mp)    = 0;
    }
    else
    {
	DN_STAT_INCR(dn_stat22);
	/*
	 * First check congestion.   Here we could implement 
	 * the random early detection (RED) if we like to.
	 */
	if (pipe->routerq_size)
	{
	    if (pipe->routerq_len >= pipe->routerq_size)
	    {
		DN_STAT_INCR(dn_stat23);
		goto drop_it;
	    }
	}
	else if (pipe->routerq_size_bytes)
	{
	    if (len + pipe->routerq_len_bytes > pipe->routerq_size_bytes)
	    {
		DN_STAT_INCR(dn_stat24);
		goto drop_it;
	    }
	}
	/*
	 * Calculate how many frame errors we will have, i.e. we need
	 * retransmit those error frames locally at link level.
	 *
	 * Note that the packet error due to a frame error is different
	 * from packet loss.  The former still consumes the link bandwidth,
	 * and the packet (even though its contents are in error) is still
	 * delivered to the destination.  Eventually, the receiver-side
	 * TCP/IP will drop the packet due to the checksum error.
	 * Therefore, we can not simply drop the packet, yet.
	 *
	 * *DN_PKT_ERROR(mp) will have 1 if the packet is in error.
	 *
	 * dn_frame_rxmits also returns, in local_rxmits, the number
	 * of link-level frame retransmisiions occur.
	 */
	*(DN_PKT_ERROR(mp)) = dn_frame_rxmits(len,
					      pipe->frame_payload_size,
					      pipe->ferror_per_million,
					      pipe->ferror_rxmit_trials,
					      &local_rxmits);
	*DN_PKT_DELAY(mp) =
	    pipe->delay_ticks + pipe->ferror_delay_ticks * local_rxmits;
	/*
	 * DN_PKT_LEN is the length of the packet.
	 * DN_PKT_XMIT_LEN is the number of bytes needed to transmit
	 * the packet including the bytest used for possible local
	 * link-level frame retransmissions.
	 */
	*DN_PKT_LEN(mp) = len;
	xmit_len = len + pipe->frame_payload_size * local_rxmits;
	*DN_PKT_XMIT_LEN(mp) = xmit_len;
	
	pipe->routerq_len++;
	if (pipe->routerq_len > pipe->router_q_hiwat)
	{
	    pipe->router_q_hiwat = pipe->routerq_len;
	}
	pipe->routerq_len_bytes += len ;
    }
    /*
     * Queue this packet in the router queue.
     */
    mp->b_next = NULL;
    if (pipe->routerq_head == NULL)
    {
	DN_STAT_INCR(dn_stat25);
	/*
	 * If there is no packet in the router queue, reset the
	 * available bandwidth to 0.  We need to wait for enough
	 * bandwidth to become accumulated.
	 *
	 * However, because the resolution of tick is coarse, 
	 * e.g. 10 ms in HP-UX 11.X, for small segments (e.g. acks),
	 * it may not be fair to wait for another tick.  So, if the
	 * segment is small, we borrow part of bandwidth from the
	 * next tick.  dn_move() will do this.
	 */
	pipe->numbytes = 0;
	pipe->routerq_head = pipe->routerq_tail = mp;
	current_ticks = ticks_since_boot;
	dn_move(pipe, current_ticks);
	/*
	 * See if we need to start the timer.
	 */
	if ((d->dn_timer_id == 0) &&
	    !(d->dn_flags & DN_CLOSING) &&
	    (pipe->pipeq_head || pipe->routerq_head))
	{
	    DN_STAT_INCR(dn_stat26);
	    d->dn_last_ticks = current_ticks;
	    d->dn_timer_id = timeout(dummynet_timeout, (caddr_t)q, 1);
	}
	else
	{
	    DN_STAT_INCR(dn_stat27);
	}
    }
    else
    {
	DN_STAT_INCR(dn_stat28);
	pipe->routerq_tail->b_next = mp;
	pipe->routerq_tail = mp;
	/*
	 * dn_timer_id should not be 0 here.
	 */
	if (d->dn_timer_id == 0)
	{
	    DN_STAT_INCR(dn_stat29);
	}
    }
    return;
    
drop_it:
    pipe->routerq_drops++;
    freemsg(mp);
    return;
}


/*
 * dn_frame_rxmits()
 *
 *    Simulate ARQ (Auto Retransmit Request) and compute how many
 *    frames we need to retransmit locally at link level.
 *
 *    Returns 1 : packet error
 *            0 : No packet error
 *
 *    Also returns, in *rxmits (the last parameter), the number of
 *    link-level frame ratransmissions that will occur.
 *
 *    Note that that the packet loss due to frame errors is different
 *    from the packet loss due to congestion.  The former still consumes
 *    the link bandwidth.
 */
static int
dn_frame_rxmits(int len,
		int frame_payload_size,
		int frame_errors_per_million,
		int frame_rxmit_max,
		int *rxmits)    /* # of frames to retransmit */
{
    int		i;
    int		j;
    int		frames;		/* # of frames per packet    */
    int		retransmits;
    int		packet_error;

    if (!frame_payload_size || !frame_errors_per_million)
    {
	*rxmits = 0;
	return 0;
    }
    frames = (len + frame_payload_size - 1)/frame_payload_size;

    retransmits = 0;
    packet_error = 0;

    for (i = 0; i < frames; i++)
    {
	for (j = 0; j <= frame_rxmit_max; j++)
	{
	    DN_STAT_INCR(dn_stat30);
	    /*
	     * dn_random() returns a random number between 0 and 1 million.
	     */
	    if (dn_random() >= frame_errors_per_million)
	    {
		DN_STAT_INCR(dn_stat31);
		/*
		 * Good.  We successfully transmit the frame.
		 */
		break;
	    }
	    /*
	     * Frame in error.
	     */
	    if (j < frame_rxmit_max)
	    {
		DN_STAT_INCR(dn_stat32);
		retransmits++;	/* local retransmission */
	    }
	    else
	    {
		/*
		 * We failed to transmit the frame correctly.
		 */
		DN_STAT_INCR(dn_stat33);
		packet_error = 1;
	    }
	}
    }
    *rxmits = retransmits;
    return packet_error;
}



/*
 * dn_move()
 *
 *    Move packets from router queue to pipe queue.
 *
 *    Dummynet has a pair of pipes: one for outbound (write-side,
 *    sending-side), and one for inbound (read-side, receiving-side).
 *    Each pipe has a router queue and a pipe queue.  The router
 *    queue controls the bandwidth of the link simulated by the
 *    dummynet.  The pipe queue controls the latency of th link
 *    simulated by the dummynet.  Each packet is first queued in
 *    the router queue, and waits for the sufficient bandwidth
 *    become available for sending the packet.  Then, the packet
 *    is moved to the pipe queue, and waits for the delay/latency
 *    time to be passed.  The delay time simulates the latency
 *    between the sender and receiver over the simulated link,
 *    and it includes the processing time of the transmiter and
 *    the receiver, the propagation delay, the processing time
 *    of link-levle frame retransmission(s), etc.
 */
static void
dn_move(dn_pipe_t *pipe, unsigned int current_ticks)
{
    queue_t		*q;
    dn_d_t		*d;
    MBLKP		pkt;
    int			len;
    int			xmit_len;
    int			bytes_available;
    int			bytes_needed;
    int			max_borrowable;
    int			delay;

    DN_STAT_INCR(dn_stat34);

    q = pipe->owner_q;
    d = (dn_d_t *) q->q_ptr;

    /*
     * Process the packets in the router queue first.  As far as there
     * is available bandwidth, move packets from the router queue
     * to the pipe queue.
     */
    if (pkt = pipe->routerq_head)
    {
	DN_STAT_INCR(dn_stat35);
	while (pkt)
	{
	    DN_STAT_INCR(dn_stat36);
	    len = *DN_PKT_LEN(pkt);
	    xmit_len = *DN_PKT_XMIT_LEN(pkt); /* Including local retransmits */
	    /*
	     * If len is 0, it's a control primitive between upstream and
	     * downstream.  So, just process it (i.e. skip the if-statement
	     * below.
	     */
	    if (len)
	    {
		DN_STAT_INCR(dn_stat37);
		if (pipe->bandwidth_bps)
		{
		    bytes_available = pipe->numbytes;
		    if (bytes_available >= xmit_len)
		    {
			DN_STAT_INCR(dn_stat38);
			pipe->numbytes -= xmit_len;
		    }
		    else
		    {
			DN_STAT_INCR(dn_stat39);
			/*
			 * Because the timeout resolution is coarse,
			 * e.g. 10 ms in HP-UX 11.X, if the packet is
			 * small in size (e.g. ack), it may not be fair
			 * to wait for another tick.  So, check if we
			 * want to borrow part of bandwidth from the next
			 * tick.  Allow, at maximum, half of the bandwidth 
			 * to be borrowed from the next tick.
			 */
			bytes_needed = xmit_len - bytes_available;
			max_borrowable = ((pipe->bandwidth_bps/8)/2)/HZ;
			if ((bytes_needed + pipe->bytes_borrowed)
			    <= max_borrowable)
			{
			    DN_STAT_INCR(dn_stat40);
			    /*
			     * Remember how much we've borrowed so far.
			     */
			    pipe->bytes_borrowed += bytes_needed;
			    pipe->numbytes = 0;
			}
			else
			{
			    DN_STAT_INCR(dn_stat41);
			    break;
			}
		    }
		}
		pipe->routerq_len--; /* elements in queue */
		pipe->routerq_len_bytes -= len;
		/*
		 * Calculate the time this packet will get out of the pipe.
		 * While the packet is in the router queue, *DN_PKT_DELAY(pkt)
		 * contains the # of ticks to delay.  When we move the packet
		 * to the pipe queue, *DN_PKT_DELAY(pkt) now contains the
		 * number of ticks since boot at which the packet will go
		 * out of the pipe.
		 */
		*DN_PKT_DELAY(pkt) = *DN_PKT_DELAY(pkt) + current_ticks;
	    }
	    /*
	     * Queue the packet in the pipe.
	     */
	    if (pipe->pipeq_head == NULL)
	    {
		DN_STAT_INCR(dn_stat42);
		pipe->pipeq_head = pkt;
	    }
	    else
	    {
		DN_STAT_INCR(dn_stat43);
		pipe->pipeq_tail->b_next = pkt;
	    }
	    pipe->pipeq_tail = pkt;
	    pkt = pkt->b_next;
	    pipe->pipeq_tail->b_next = NULL;
	}
	pipe->routerq_head = pkt;
	/*
	 * Just a sanity check.  The routerq_len should be zero if there
	 * is no more packet in the router queue.
	 */
	if (pkt == NULL && pipe->routerq_len != 0)
	{
	    DN_STAT_INCR(dn_stat44);
	}
    }
    /*
     * Deliver packets after the delay in the pipe queue.
     */
    while ((pkt = pipe->pipeq_head) &&
	   (*DN_PKT_DELAY(pkt)) <= ticks_since_boot)
    {
	DN_STAT_INCR(dn_stat45);
	pipe->pipeq_head = pkt->b_next;
	pkt->b_next = 0;
	if (*DN_PKT_ERROR(pkt))
	{
	    /*
	     * Packet error.  Just discard this.
	     */
	    DN_STAT_INCR(dn_stat46);
	    freemsg(pkt);
	}
	else
	{
	    DN_STAT_INCR(dn_stat47);
	    putnext(q, pkt);
	}
    }
}


/*
 * dummynet_timeout()
 *
 *    We specify this routine to be called back when our timer pops up.
 *
 *    In order to safely work on the dummynet data structures, we need
 *    some synchronization mechanism.  In HP-UX 11.X, we use the
 *    STREAMS-provided synchronization mechanism: SQLVL_QUEUEPAIR,
 *    which means only one thread can execute put/service routine
 *    of this dummynet instance at a time.  In order to obtain this
 *    synchronization protection from outside the STREAMS framework
 *    (i.e., in our case, from timeout's callback routine), we need
 *    to call streams_put() routine.  The streams_put() will do the
 *    necessary things for us to obtain the synchronization and call
 *    our real timeout routine: dummynet_timeout_int().
 */
static void
dummynet_timeout(caddr_t arg)
{
    queue_t	*q;
    dn_d_t	*d;
    mblk_t	*mp;

    DN_STAT_INCR(dn_stat48);

    q = (queue_t *)arg;
    d = (dn_d_t *) q->q_ptr;
    /*
     * We need to pass an mblk to use streams_put().  We use the
     * mblk that was pre-allocated at dummynetopen().
     *
     * Because the mblk passed to the streams_put() can
     * be freed by STREAMS framework if the dummynet is
     * popped (i.e. closed) while this callback request is
     * still outstanding, until dummynet_imeout_int() is actually
     * called back, we are not the owner of the dn_timer_mp.
     */
    mp = d->dn_timer_mp;
    d->dn_timer_mp = 0;
    streams_put((streams_put_t)dummynet_timeout_int, q, mp, q);
}

/*
 * dummynet_timeout_int()
 *
 *    Do all those things required on every tick.
 */
static void
dummynet_timeout_int(queue_t *q, mblk_t *mp)
{
    dn_d_t	 	*d;
    dn_pipe_t		*pipe_outbound;
    dn_pipe_t		*pipe_inbound;
    int			s;
    int			bandwidth;
    MBLKP		pkt;
    unsigned int	current_ticks;
    unsigned int	ticks_passed;
    int			available_bandwidth;
    int			newbytes;
    
    DN_STAT_INCR(dn_stat49);

    d = (dn_d_t *) q->q_ptr;
    /*
     * Now, we can safely restore the mblk that was used for
     * calling streams_put() back to dn_timer_mp.
     */
    d->dn_timer_mp = mp;
    /*
     * Figure out how many ticks have passed since the last time.
     */
    current_ticks = ticks_since_boot;
    ticks_passed = current_ticks - d->dn_last_ticks;
    d->dn_last_ticks = current_ticks;

    if (ticks_passed == 1)
	DN_STAT_INCR(dn_stat50);
    else if (ticks_passed < 1)
	DN_STAT_INCR(dn_stat51);
    else
	DN_STAT_INCR(dn_stat52);
    /*
     * Process the outbound/write side.
     */
    pipe_outbound = &d->dn_pipe_outbound;
    available_bandwidth = ((pipe_outbound->bandwidth_bps/8)/HZ) * ticks_passed;
    dn_compute_numbytes(pipe_outbound, available_bandwidth);
    dn_move(pipe_outbound, current_ticks);
    /*
     * Process the inbound/read side.
     */
    pipe_inbound = &d->dn_pipe_inbound;
    available_bandwidth = ((pipe_inbound->bandwidth_bps/8)/HZ) * ticks_passed;
    dn_compute_numbytes(pipe_inbound, available_bandwidth);
    dn_move(pipe_inbound, current_ticks);
    /*
     * If some queue has still data and if the module is not closing,
     * restart the timer.
     */
    if (!(d->dn_flags & DN_CLOSING) &&
	(pipe_outbound->pipeq_head || pipe_outbound->routerq_head ||
	 pipe_inbound->pipeq_head  || pipe_inbound->routerq_head))
    {
	DN_STAT_INCR(dn_stat53);
	d->dn_timer_id = timeout(dummynet_timeout, (caddr_t) q, 1);
    }
    else
    {
	DN_STAT_INCR(dn_stat54);
	d->dn_timer_id = 0;
    }
}


static void
dn_compute_numbytes(dn_pipe_t *pipe, int available_bandwidth)
{
    int		newbytes;
    
    if (pipe->routerq_head)
    {
	DN_STAT_INCR(dn_stat55);
	newbytes = available_bandwidth - pipe->bytes_borrowed;
	if (newbytes > 0)
	{
	    DN_STAT_INCR(dn_stat56);
	    pipe->numbytes += newbytes;
	    pipe->bytes_borrowed = 0;
	}
	else
	{
	    DN_STAT_INCR(dn_stat57);
	    pipe->bytes_borrowed -= available_bandwidth;
	}
    }
    else
    {
	DN_STAT_INCR(dn_stat58);
	pipe->numbytes = 0;
	pipe->bytes_borrowed = 0;
    }
}



/*
 * dummynet_flush_putnext()
 *
 *    Flush (not delete) all the packets to the next module.
 */
static void
dummynet_flush_putnext(dn_pipe_t *pipe)
{
    MBLKP	pkt;
    queue_t	*q;

    DN_STAT_INCR(dn_stat59);

    q = pipe->owner_q;

    while (pipe->pipeq_head)
    {
	DN_STAT_INCR(dn_stat60);
	pkt = pipe->pipeq_head;
	pipe->pipeq_head = pkt->b_next;
	pkt->b_next = 0;
	putnext(q, pkt);
    }
    while (pipe->routerq_head)
    {
	DN_STAT_INCR(dn_stat61);
	pkt = pipe->routerq_head;
	pipe->routerq_head = pkt->b_next;
	pkt->b_next = 0;
	putnext(q, pkt);
    }
    pipe->routerq_len = 0;
    pipe->routerq_len_bytes = 0;
}


/*
 * dummynet_flush_freemsg()
 *
 *    Flush and free all the packets.
 */
static void
dummynet_flush_freemsg(dn_pipe_t *pipe)
{
    MBLKP	pkt;

    DN_STAT_INCR(dn_stat62);

    while (pipe->pipeq_head)
    {
	DN_STAT_INCR(dn_stat63);
	pkt = pipe->pipeq_head;
	pipe->pipeq_head = pkt->b_next;
	pkt->b_next = 0;
	freemsg(pkt);
    }
    while (pipe->routerq_head)
    {
	DN_STAT_INCR(dn_stat64);
	pkt = pipe->routerq_head;
	pipe->routerq_head = pkt->b_next;
	pkt->b_next = 0;
	freemsg(pkt);
    }
    pipe->routerq_len = 0;
    pipe->routerq_len_bytes = 0;
}


/* 
 * dn_random()
 * 
 *    Generates a uniformly distributed random number between 0 and
 *    1,000,000.
 *
 * Reference: Raj Jain's book on perf. evaluation and measurement.
 */
static int seed = 1;

#define DN_RANDOM_START	0
#define DN_RANDOM_RANGE	1000000
#define DN_RANDOM_A	16807		/* Multiplier */
#define DN_RANDOM_M	MAXINT	       	/* Modulus */
#define DN_RANDOM_Q	127773		/* m / a */
#define DN_RANDOM_R	2836		/* m % a */

static int 
dn_random()
{
    int		x_div_q;
    int		x_mod_q;
    float	rn;

    x_div_q = seed / DN_RANDOM_Q;
    x_mod_q = seed % DN_RANDOM_Q;
    seed = DN_RANDOM_A * x_mod_q - DN_RANDOM_RANGE * x_div_q;

    /* make seed positive by adding MAXINT if necessary */
    if (seed <= 0)
	seed += DN_RANDOM_M;

    /* rn ranges between 0 and 1 */
    rn = (float)seed/(float)DN_RANDOM_M;
    return((int)(rn * DN_RANDOM_RANGE + DN_RANDOM_START));
}

/*
 * dummynet_ioctl
 *
 *    Processes M_IOCTL message.
 *
 *    Returns 1 if the ioctl command is for dummynet.
 *    Returns 0 otherwise.
 */
static int
dummynet_ioctl(queue_t *q, mblk_t *mp)
{
    struct iocblk	*iocp;
    mblk_t		*mp1;
    dn_option_t		*dnopt;
    struct copyreq	*copyreqp;
    unsigned int	data_size;
    int			n;
    int			i;
    int			r;
    int			optname;
    int			optvalue;
    
    DN_STAT_INCR(dn_stat65);

    iocp = (struct iocblk *) mp->b_rptr;

    switch (iocp->ioc_cmd)
    {
	case DN_SET_DEFAULTS:
	{
	    mp1 = mp->b_cont;
	    if (!mp1)
	    {
		mp->b_datap->db_type = M_IOCNAK;
		iocp->ioc_error = EINVAL;
		qreply(q, mp);
		return 1;
	    }
	    dnopt = (dn_option_t *) mp1->b_rptr;
	    data_size = mp1->b_wptr - mp1->b_rptr;
	    n = data_size / sizeof(dn_option_t);

	    r = n; /* return value: # of options successfully updated. */

	    for (i = 0; i < n; i++)
	    {
		optname = dnopt[i].name;
		optvalue = dnopt[i].value;

		switch (optname)
		{
		    /*
		     * General options.
		     */
		    case DN_PLACEMENT:
			switch (optvalue)
			{
			    case PUSH_DN_BETWEEN_IP_AND_DRIVER:
				dn_placement = optvalue;
				break;
			    case PUSH_DN_BETWEEN_TCP_UDP_AND_IP:
				dn_placement = optvalue;
				break;
			    default:
				/* Invalid value.  Ignore it. */
				r--;
				break;
			}
			break;
		    /*
		     * Outbound options.
		     */
		    case DN_OUTBOUND_ENABLE:
			if (optvalue > 0)
			{
			    dn_outbound_enable = 1;
			}
			else
			{
			    dn_outbound_enable = 0;
			}
			break;
		    case DN_OUTBOUND_BANDWIDTH_BPS:
			dn_outbound_bandwidth_bps = optvalue;
			break;
		    case DN_OUTBOUND_DELAY_TICKS:
			dn_outbound_delay_ticks = optvalue;
			break;
		    case DN_OUTBOUND_QUEUE_SIZE:
			dn_outbound_queue_size = optvalue;
			break;
		    case DN_OUTBOUND_QUEUE_SIZE_BYTES:
			dn_outbound_queue_size_bytes = optvalue;
			break;
		    case DN_OUTBOUND_FRAME_PAYLOAD_SIZE:
			dn_outbound_frame_payload_size = optvalue;
			break;
		    case DN_OUTBOUND_FERROR_PER_MILLION:
			dn_outbound_ferror_per_million = optvalue;
			break;
		    case DN_OUTBOUND_FERROR_DELAY_TICKS:
			dn_outbound_ferror_delay_ticks = optvalue;
			break;
		    case DN_OUTBOUND_FERROR_RXMIT_TRIALS:
			dn_outbound_ferror_rxmit_trials = optvalue;
			break;
		    /*
		     * Inbound options.
		     */
		    case DN_INBOUND_ENABLE:
			if (optvalue > 0)
			{
			    dn_inbound_enable = 1;
			}
			else
			{
			    dn_inbound_enable = 0;
			}
			break;
		    case DN_INBOUND_BANDWIDTH_BPS:
			dn_inbound_bandwidth_bps = optvalue;
			break;
		    case DN_INBOUND_DELAY_TICKS:
			dn_inbound_delay_ticks = optvalue;
			break;
		    case DN_INBOUND_QUEUE_SIZE:
			dn_inbound_queue_size = optvalue;
			break;
		    case DN_INBOUND_QUEUE_SIZE_BYTES:
			dn_inbound_queue_size_bytes = optvalue;
			break;
		    case DN_INBOUND_FRAME_PAYLOAD_SIZE:
			dn_inbound_frame_payload_size = optvalue;
			break;
		    case DN_INBOUND_FERROR_PER_MILLION:
			dn_inbound_ferror_per_million = optvalue;
			break;
		    case DN_INBOUND_FERROR_DELAY_TICKS:
			dn_inbound_ferror_delay_ticks = optvalue;
			break;
		    case DN_INBOUND_FERROR_RXMIT_TRIALS:
			dn_inbound_ferror_rxmit_trials = optvalue;
			break;

		    default:
			/* Unknown option.  Ignore it. */
			r--;
			break;
		} /* switch (optname) */
	    } /* for */
	    
	    mp->b_datap->db_type = M_IOCACK;
	    iocp->ioc_error = 0;
	    iocp->ioc_count = 0;
	    iocp->ioc_rval  = r;
	    freemsg(mp1);
	    mp->b_cont = NULL;
	    qreply(q, mp);
	    return 1;
	}
	case DN_GET_ALL_DEFAULTS:
	{
	    mp1 = mp->b_cont;
	    if (!mp1)
	    {
		DN_STAT_INCR(dn_stat66);
		mp->b_datap->db_type = M_IOCNAK;
		iocp->ioc_error = EINVAL;
		qreply(q, mp);
		return 1;
	    }
	    data_size = DN_NUMBER_OF_OPTIONS * sizeof(dn_option_t);
	    if ((mp1->b_wptr - mp1->b_rptr) < data_size)
	    {
		DN_STAT_INCR(dn_stat67);
		freemsg(mp1);
		mp->b_cont = NULL;
		mp->b_datap->db_type = M_IOCNAK;
		iocp->ioc_error = EINVAL;
		qreply(q, mp);
		return 1;
	    }
	    DN_STAT_INCR(dn_stat68);

	    mp->b_datap->db_type = M_IOCACK;
	    iocp->ioc_error = 0;
	    iocp->ioc_count = data_size;
	    iocp->ioc_rval  = 0;

	    dnopt = (dn_option_t *) mp1->b_rptr;

	    dnopt[0].name   = DN_PLACEMENT;
	    dnopt[0].value  = dn_placement;
	    dnopt[1].name   = DN_OUTBOUND_ENABLE;
	    dnopt[1].value  = dn_outbound_enable;
	    dnopt[2].name   = DN_OUTBOUND_BANDWIDTH_BPS;
	    dnopt[2].value  = dn_outbound_bandwidth_bps;
	    dnopt[3].name   = DN_OUTBOUND_DELAY_TICKS;
	    dnopt[3].value  = dn_outbound_delay_ticks;
	    dnopt[4].name   = DN_OUTBOUND_QUEUE_SIZE;
	    dnopt[4].value  = dn_outbound_queue_size;
	    dnopt[5].name   = DN_OUTBOUND_QUEUE_SIZE_BYTES;
	    dnopt[5].value  = dn_outbound_queue_size_bytes;
	    dnopt[6].name   = DN_OUTBOUND_FRAME_PAYLOAD_SIZE;
	    dnopt[6].value  = dn_outbound_frame_payload_size;
	    dnopt[7].name   = DN_OUTBOUND_FERROR_PER_MILLION;
	    dnopt[7].value  = dn_outbound_ferror_per_million;
	    dnopt[8].name   = DN_OUTBOUND_FERROR_DELAY_TICKS;
	    dnopt[8].value  = dn_outbound_ferror_delay_ticks;
	    dnopt[9].name   = DN_OUTBOUND_FERROR_RXMIT_TRIALS;
	    dnopt[9].value  = dn_outbound_ferror_rxmit_trials;
	    dnopt[10].name  = DN_INBOUND_ENABLE;
	    dnopt[10].value = dn_inbound_enable;
	    dnopt[11].name  = DN_INBOUND_BANDWIDTH_BPS;
	    dnopt[11].value = dn_inbound_bandwidth_bps;
	    dnopt[12].name  = DN_INBOUND_DELAY_TICKS;
	    dnopt[12].value = dn_inbound_delay_ticks;
	    dnopt[13].name  = DN_INBOUND_QUEUE_SIZE;
	    dnopt[13].value = dn_inbound_queue_size;
	    dnopt[14].name  = DN_INBOUND_QUEUE_SIZE_BYTES;
	    dnopt[14].value = dn_inbound_queue_size_bytes;
	    dnopt[15].name  = DN_INBOUND_FRAME_PAYLOAD_SIZE;
	    dnopt[15].value = dn_inbound_frame_payload_size;
	    dnopt[16].name  = DN_INBOUND_FERROR_PER_MILLION;
	    dnopt[16].value = dn_inbound_ferror_per_million;
	    dnopt[17].name  = DN_INBOUND_FERROR_DELAY_TICKS;
	    dnopt[17].value = dn_inbound_ferror_delay_ticks;
	    dnopt[18].name  = DN_INBOUND_FERROR_RXMIT_TRIALS;
	    dnopt[18].value = dn_inbound_ferror_rxmit_trials;
	    mp1->b_wptr = (unsigned char *)&dnopt[19];
	    
	    qreply(q, mp);

	    return 1;
	}
	default:
	{
	    DN_STAT_INCR(dn_stat69);
	    return 0;
	}
    }
}

    

