/* 
 * dummynetopt.h
 *
 *    This file contains dummynet's ioctl commands and their associated
 *    data structs.
 *
 *
 *            Copyright (C) 1999 Hewlett-Packard Company
 *                       ALL RIGHTS RESERVED.
 *
 * The enclosed software and documentation includes copyrighted works
 * of Hewlett-Packard Co. For as long as you comply with the following
 * limitations, you are hereby authorized to (i) use, reproduce, and
 * modify the software and documentation, and to (ii) distribute the
 * software and documentation, including modifications, for
 * non-commercial purposes only.
 *      
 * 1.  The enclosed software and documentation is made available at no
 *     charge in order to advance the general development of
 *     high-performance networking products.
 * 
 * 2.  You may not delete any copyright notices contained in the
 *     software or documentation. All hard copies, and copies in
 *     source code or object code form, of the software or
 *     documentation (including modifications) must contain at least
 *     one of the copyright notices.
 * 
 * 3.  The enclosed software and documentation has not been subjected
 *     to testing and quality control and is not a Hewlett-Packard Co.
 *     product. At a future time, Hewlett-Packard Co. may or may not
 *     offer a version of the software and documentation as a product.
 * 
 * 4.  THE SOFTWARE AND DOCUMENTATION IS PROVIDED "AS IS".
 *     HEWLETT-PACKARD COMPANY DOES NOT WARRANT THAT THE USE,
 *     REPRODUCTION, MODIFICATION OR DISTRIBUTION OF THE SOFTWARE OR
 *     DOCUMENTATION WILL NOT INFRINGE A THIRD PARTY'S INTELLECTUAL
 *     PROPERTY RIGHTS. HP DOES NOT WARRANT THAT THE SOFTWARE OR
 *     DOCUMENTATION IS ERROR FREE. HP DISCLAIMS ALL WARRANTIES,
 *     EXPRESS AND IMPLIED, WITH REGARD TO THE SOFTWARE AND THE
 *     DOCUMENTATION. HP SPECIFICALLY DISCLAIMS ALL WARRANTIES OF
 *     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * 5.  HEWLETT-PACKARD COMPANY WILL NOT IN ANY EVENT BE LIABLE FOR ANY
 *     DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
 *     (INCLUDING LOST PROFITS) RELATED TO ANY USE, REPRODUCTION,
 *     MODIFICATION, OR DISTRIBUTION OF THE SOFTWARE OR DOCUMENTATION.
 */

#ifndef _DUMMYNETOPT_H_
#define _DUMMYNETOPT_H_

#ifdef _KERNEL_BUILD
#  include "../h/stropts.h"
#else
#  include <sys/stropts.h>
#endif /* _KERNEL_BUILD */

/*
 * Dummynet ioctl commands.
 *
 *    DN_SET_DEFAULTS		Set dummynet's default setting(s).  The change
 *                              will be effecitve for the subsequently opened
 *                              dummynets, and will not affect the dummynets
 *             			that have already been opened.
 *                              
 *    DN_GET_ALL_DEFAULTS	Get all the dummynet's default settings.
 */
#define DNIOC			'd'
#define DN_SET_DEFAULTS		_IO(DNIOC, 1)
#define DN_GET_ALL_DEFAULTS	_IO(DNIOC, 2)

/*
 * Dummynet options.
 */
struct dn_option
{
	int	name;	/* option name */
	int	value;  /* value       */
};
typedef struct dn_option dn_option_t;

/*
 * General options.
 */
#define DN_PLACEMENT			0x00 /* Where ummynet is pushed.
						Possible option values are
					        the following.               */
#  define PUSH_DN_BETWEEN_IP_AND_DRIVER	0    /*   o between IP and driver    */
#  define PUSH_DN_BETWEEN_TCP_AND_IP   	1    /*   o between TCP and IP       */
/*
 * Outbound options.
 */
#define DN_OUTBOUND_ENABLE		0x10 /* enable/disable dummynet      */
#define DN_OUTBOUND_BANDWIDTH_BPS	0x11 /* bits per second              */
#define DN_OUTBOUND_DELAY_TICKS		0x12 /* pipe latency in ticks        */
#define DN_OUTBOUND_QUEUE_SIZE		0x13 /* q size in # pkts: 1st choice */
#define DN_OUTBOUND_QUEUE_SIZE_BYTES	0x14 /* q size in bytes:  2nd choice */
#define DN_OUTBOUND_FRAME_PAYLOAD_SIZE	0x15 /* frame size in bytes	     */
#define DN_OUTBOUND_FERROR_PER_MILLION	0x16 /* frame errors per 10^6 frames */
#define DN_OUTBOUND_FERROR_DELAY_TICKS	0x17 /* cost of a local frame rexmit */
#define DN_OUTBOUND_FERROR_RXMIT_TRIALS	0x18 /* # attempts of local rexmits
                                                excluding original transmit  */
/*
 * Inbound options.
 */
#define DN_INBOUND_ENABLE		0x20 /* enable/disable dummynet      */
#define DN_INBOUND_BANDWIDTH_BPS	0x21 /* bits per second              */
#define DN_INBOUND_DELAY_TICKS		0x22 /* pipe latency in ticks        */
#define DN_INBOUND_QUEUE_SIZE		0x23 /* q size in # pkts: 1st choice */
#define DN_INBOUND_QUEUE_SIZE_BYTES	0x24 /* q size in bytes:  2nd choice */
#define DN_INBOUND_FRAME_PAYLOAD_SIZE	0x25 /* frame size in bytes	     */
#define DN_INBOUND_FERROR_PER_MILLION	0x26 /* frame errors per 10^6 frames */
#define DN_INBOUND_FERROR_DELAY_TICKS	0x27 /* cost of a local frame rexmit */
#define DN_INBOUND_FERROR_RXMIT_TRIALS	0x28 /* # attempts of local rexmits
                                                excluding original transmit  */
/*
 * The number of options that DN_GET_ALL_DEFAULTS may return.
 */
#define DN_NUMBER_OF_OPTIONS		19

#endif /* _DUMMYNETOPT_H_ */
